function out = catLGM(model, task, algo, data, varargin)
% Categorical data analysis using Latent Gaussian Models
% All function calls follow the following format to performa specific task on a
% data given a model and an algorithm. 
%
% out = catLGM(model, task, algo, data)
%
% For example, given features X and labels y, the following performs posterior
% inference in Bayesian logistic regression using a variational method based on
% the piecewise bound, and then predict labels for new test data using a 
% Monte-carlo estimate.  
%
% postDist = catLGM('bayesLogitReg', 'infer', 'var-pw', [X y]);
% pred = catLGM('bayesLogitReg', 'predict', 'mc', [Xtest], [], postDist);
%
% Similarly, given a categorical data matrix Y with missing values, the following
% learns a Multinomial-logit factor analysis model using a variational EM
% algorithm based on the log bound, and then imputes the missing values using the
% learned model (with a Monte-carlo estimate).
%
% out = catLGM('multiLogitFA', 'infer', 'var-log', Y);
% pred = catLGM('multiLogitFA', 'impute', 'mc', Y, [], out.params, out.postDist);
% 
% See full list of model-task-algo pair below in the code.
%
% Algorithmic options can be specified as 5'th argument
% options = struct('lowerBoundTolLearn', 0.01,'displayLearn',0);
% out = catLGM('multiLogitFA', 'infer', 'var-log', Y, options);
% See first few lines below for a list of options
%
% Written by Emtiyaz, CS, UBC
% Modified on Dec. 16, 2011

  optargin = size(varargin,2);
  if optargin==0
    options = [];
    input1 = [];
  elseif optargin==1 
    options = varargin{1};
    input1 = [];
  elseif optargin==2 
    options = varargin{1};
    input1 = varargin{2}; % can be params/ postDist depending on task
  elseif optargin==3 
    options = varargin{1};
    input1 = varargin{2}; % params
    input2 = varargin{3}; % postDist depending on task
  else
      error('too many input arguments');
  end

  % get default algorithmic options
  [maxItersLearn, lowerBoundTolLearn, displayLearn, computeSs,...
	maxItersInfer, lowerBoundTolInfer, initVec, displayInfer,...
        maxItersMstep, lowerBoundTolMstep, displayMstep, ...
	boundParams] = getCatLGMDefaultOptions(options);

  % set Inference options 
  options.maxItersInfer = maxItersInfer;
  options.lowerBoundTol = lowerBoundTolInfer;
  options.display = displayInfer;

  % set Mstep options
  options.maxItersMstep = maxItersMstep;
  options.lowerBoundTolMstep = lowerBoundTolMstep;
  options.displayMstep = displayMstep;

  % initialize bound parameters
  options.boundParams = boundParams;

  % models
  switch model
  case {'bayesLogitReg','bayesMultiLogitReg'}
    switch task
    case 'infer'
      [data, params] = setupLGM(data, model, options);
      [inferFunc, boundFunc, options] =...
          getAlgoFuncsAndOptions(model, task, algo, options, [], params);
      [tt1, logLik, postDist] = inferFunc(data, params, boundFunc, options);
      out = postDist;
      out.logLik = logLik;
      out.nClass = unique(params.others.nClass);

    case 'predict'
      % get postDist
      postDist = input1;
      [D,F] = size(data); 
      K = input1.nClass;
      M = K-1;

      % factor loading matrix and bias
      X = data;
      W = zeros(D*(M+1),F*M);
      for k = 1:K-1
        idx1 = k:K:(D-1)*K+k;
        idx2 = F*(k-1)+1:F*k;
        W(idx1,idx2) = X;
      end
      w0 = zeros(size(W,1),1);
      regBias = postDist.regBias;
      bias = w0 + repmat([regBias;0],D,1);

      switch algo
      case 'mc'
        % monte-carlo estimate to predictive prob
        % p(y=k) = int_z exp(eta_k-lse(eta))*N(m,V) dz
        S = 10000;
        z = mvnrnd(postDist.mean', postDist.covMat, S);
        eta = bsxfun(@plus, W*z', bias);
        eta = reshape(eta,K,D*S);
        lse = logsumexp(eta,1);
        prob = exp(bsxfun(@minus, eta, lse));
        prob = reshape(prob,K*D,S);
        prob = reshape(mean(prob,2),K,D);
        out = prob';
      otherwise
        error('no such algo');
      end

    otherwise
      error('no such task');
    end

  case {'binaryLogitLGGM', 'binaryLogitFA', 'multiLogitLGGM','multiLogitFA','stickLogitLGGM','stickLogitFA'}
  
    % inference algorithm function
    switch task
    case {'infer'}
      params = input1;
      [data, junk] = setupLGM(data, model, params.others);
      [inferFunc, boundFunc, options] = getAlgoFuncsAndOptions(model, task, algo, options, [], params);
      [tt1, logLik, postDist] = inferFunc(data, params, boundFunc, options);
      postDist.mean = postDist.mean(:,data.mappingJ);
      postDist.covMat = postDist.covMat(:,:,data.mappingJ);
      postDist.logLik_n = postDist.logLik_n(:,data.mappingJ);
      postDist.logLik = logLik;

    case {'impute'}
      % get params and postDist
      params = input1;
      postDist = input2;
      nC = params.others.nClass;
      meanPost = postDist.mean;
      covMatPost = postDist.covMat;

      % data
      Y = data;
      [D, N] = size(Y);
      Ybitvec = encodeDataOneOfM(Y, nC, 'M');

      switch algo
      case 'mc'
        for n = 1:N
          % sample latent vector
          V = covMatPost(:,:,n);
          V = (V+V')/2;
          z = mvnrnd(meanPost(:,n)', V, 10000);
          didx = find(isnan(Y(:,n)));
          % for all missing dimensions
          for d = didx(:)'
            % compute prdictor eta
            idx = sum(nC(1:d-1))+1:sum(nC(1:d));
            Wd = params.loadingMat(idx,:);
            w0d = params.bias(idx,:);
            eta_d = bsxfun(@plus, Wd*z', w0d);
            % pass eta through the likelihood
            switch model 
            case {'multiLogitLGGM','multiLogitFA','binaryLogitLGGM','binaryLogitFA'}
              lse = logsumexp(eta_d,1);
              prob = exp(bsxfun(@minus, eta_d, lse));
            case {'stickLogitLGGM','stickLogitFA'}
              sig = sigmoid(eta_d);
              prob = zeros(size(sig));
              for m = 1:size(prob,1)-1
                prob(m,:) = prod([(1-sig(1:m-1,:)); sig(m,:)], 1);
              end
              prob(end,:) = prod(1-sig(1:nC(d)-1,:), 1);
            otherwise
              error('no such method');
            end
            % estimate predictive prob
            phat = mean(prob,2);
            Ybitvec(idx,n) = phat;
          end
        end
        out = Ybitvec;

      otherwise
        error('no such algo');
      end

    case 'learn'
      % set up LGM
      [data, params0] = setupLGM(data, model, options);

      % initialize
      if isempty(input1); 
        params = params0;
      else
        params = input1;
      end

      % algorithm options and functions
      [inferFunc, boundFunc, options, mstepFunc] = getAlgoFuncsAndOptions(model, task, algo, options, [], params);

      % run EM algorithm
      tic;
      for iter = 1:maxItersLearn
        [ss, logLik(iter), postDist] = inferFunc(data, params, boundFunc, options);

          [converged, incr] = isConverged(logLik, lowerBoundTolLearn, 'objFun');
          if displayLearn; fprintf('iter %d logLik %.3f\n', iter, logLik(iter)); end;
          if converged; break; end;
          if incr<-1e-10; warning('Lower bound decreased'); end

        [params, options] = mstepFunc(ss, data, params, boundFunc, options);
        tt(iter) = toc;
      end

      % output
      out.params = params;
      out.postDist.mean = postDist.mean(:,data.mappingJ);
      out.postDist.covMat = postDist.covMat(:,:,data.mappingJ);
      out.postDist.logLik_n = postDist.logLik_n(:,data.mappingJ);
      out.logLik = logLik;
      out.ss = ss;
      out.time = tt;

    otherwise
      error('no such task');
    end

  case {'logitGP', 'multiLogitGP','stickLogitGP'}
    %TODO implement learning hyperparameters
    switch task
    case {'infer'}
      [data, params0] = setupLGM(data, model, options);
      if isempty(input1)
        params = params0;
      else
        params = input1;
      end
      [inferFunc, boundFunc, options] = getAlgoFuncsAndOptions(model, task, algo, options, data, params);
      % infer
      [post, nLogLik] = inferFunc(params.hyp, params.meanfunc, params.covfunc, data.X, data.ybitvec, params.others.nClass, options);
      % output
      out.params = params;
      out.Xtrain = data.X;
      out.postDist = post;
      out.logLik = -nLogLik';

    case 'predict'
      % get postDist
      if isempty(input1);
        error('specify postDist, Xtrain, and params');
      else
        params = input1.params;
        postDist = input1.postDist;
        Xtrain = input1.Xtrain;
      end

      % get parameters
      nClass = params.others.nClass;
      hyp = params.hyp;
      covfunc = params.covfunc;
      meanfunc = params.meanfunc;

      % joint distribution of all z* and z for prediction
      xs = data;
      x = Xtrain;
      Sigma_s = covFuncMultiClass(covfunc, nClass, hyp.cov, xs, x);
      Sigma_ss = covFuncMultiClass(covfunc, nClass, hyp.cov, xs, 'diag');
      mu_s = meanFuncMultiClass(meanfunc, nClass, hyp.mean, xs);

      % prediction
      switch algo
      case 'mc'
        S = 10000;
        switch model
        case {'logitGP','multiLogitGP'}
          phat = predProbMultiLogitGP(mu_s, Sigma_ss, Sigma_s, postDist, nClass, S);
        case {'stickLogitGP'}
          phat = predProbStickLogitGP(mu_s, Sigma_ss, Sigma_s, postDist, nClass, S);
        end
        out = phat';
      otherwise
        error('no such algo');
      end

    otherwise
      error('no such task');
    end

  otherwise
    error('no such model');
  end


