function [inferFunc, boundFunc, options, mstepFunc] = getAlgoFuncsAndOptions(model, task, algo, options, data, params)
% get algorithm functions and options for a specified model, task and algorithm
  [nPieces] = myProcessOptions(options, 'nPieces', 5);

  % inference algorithm function
  switch model
  case {'bayesLogitReg','bayesMultiLogitReg'}
    K = unique(params.others.nClass);
    switch algo
    case 'var-log'
      inferFunc = @inferMultiLogitLGM_gradient;
      boundFunc = @funObj_multiLogitLGM_log;
      options.initVec = [params.mean; packcovariance(params.covMat); zeros(K-1,1)]; % add reg bias
    case 'var-boh'
      %inferFunc = @inferMultiLogitLGM_bohning;
      inferFunc = @inferMultiLogitReg_bohning;
      boundFunc = [];
      if isempty(options.boundParams)
        nC = params.others.nClass-1;
        [nCu, inC, jnC] = unique(nC);
        for d = 1:length(nCu)
          A{d} = (eye(nCu(d))- ones(nCu(d),nCu(d))/(nCu(d) + 1))/2;
          iA{d} = inv(A{d});
        end
        options.boundParams.A = A;
        options.boundParams.iA = iA;
      end

    case 'var-pw'
      inferFunc = @inferStickLogitLGM_pw;
      boundFunc = @funObj_stickLogitLGM_pw;
      options.initVec = [params.mean; packcovariance(params.covMat); zeros(K-1,1)]; % add reg bias
      if isempty(options.boundParams)
        options.boundParams = getPiecewiseBound('quad',nPieces);
      end
    otherwise
      error('no such algo');
    end

  case {'binaryLogitLGGM', 'binaryLogitFA', 'multiLogitLGGM','multiLogitFA','stickLogitLGGM','stickLogitFA'}

    % set EM algorithm options for learning
    switch task
      case 'learn'
      options.display = 0; % no display in E and M steps
      % M step options
      options.computeSs = 1;
      switch model
      case {'multiLogitLGGM','binaryLogitLGGM','stickLogitLGGM'}
        options.estimateMean = 1;
        options.estimateCovMat = 1;
        options.estimateLoadingMat = 0;
        options.estimateBias = 0;
      case {'multiLogitFA','binaryLogitFA','stickLogitFA'}
        options.estimateMean = 0;
        options.estimateCovMat = 0;
        options.estimateLoadingMat = 1;
        options.estimateBias = 1;
      end
    end

    % algorithm functions
    switch algo
    case 'var-log'
      switch model
      case {'multiLogitLGGM','binaryLogitLGGM','multiLogitFA','binaryLogitFA'}
        inferFunc = @inferMultiLogitLGM_gradient;
        mstepFunc = @maxParamsMultiLogitLGM_gradient;
        boundFunc = @funObj_multiLogitLGM_log;
      otherwise
        error('no such algo for the requested model');
      end
    case 'var-boh'
      switch model 
      case {'multiLogitLGGM','binaryLogitLGGM','multiLogitFA','binaryLogitFA'}
        inferFunc = @inferMultiLogitLGM_bohning;
        mstepFunc = @maxParamsMultiLogitLGM_bohning;
        boundFunc = [];
        if isempty(options.boundParams)
          nC = params.others.nClass-1;
          [nCu, inC, jnC] = unique(nC);
          for d = 1:length(nCu)
            A{d} = (eye(nCu(d))- ones(nCu(d),nCu(d))/(nCu(d) + 1))/2;
            iA{d} = inv(A{d});
          end
          options.boundParams.A = A;
          options.boundParams.iA = iA;
        end
      otherwise
        error('no such algo for the requested model');
      end
    case 'var-pw'
      switch model
      case {'binaryLogitLGGM','binaryLogitFA','stickLogitFA','stickLogitLGGM'}
        inferFunc = @inferStickLogitLGM_pw;
        mstepFunc = @maxParamsStickLogitLGM_pw;
        boundFunc = [];
        if isempty(options.boundParams)
          options.boundParams = getPiecewiseBound('quad',nPieces);
        end
      otherwise
        error('no such algo for the requested model');
      end
    otherwise
      error('no such algo');
    end

  case 'stickLogitGP'
    switch algo
    case 'var-pw'  
      inferFunc = @inferStickLogitGP_pw;
      options.ind = data.ind;
      if isempty(options.boundParams)
        options.boundParams = getPiecewiseBound('quad',nPieces);
      end
    otherwise 
      error('no such algo');
    end
    boundFunc = [];

  case 'multiLogitGP'
    switch algo
    case 'var-log'
      inferFunc = @inferMultiLogitGP_log;
    case 'var-boh'
      inferFunc = @inferMultiLogitGP_bohning;
      if isempty(options.boundParams)
        % form sA matrix (square root of A)
        M = params.others.nClass-1;
        N = size(data.X,1);
        sA = sparse([]);
        A1 = (eye(M) - ones(M,M)/(M + 1))/2;
        sA1 = chol(A1);
        for n = 1:N
          sA = blkdiag(sA,sA1);
        end
        options.boundParams.sA = sA;
        options.boundParams.A = sA'*sA;
      end
    otherwise 
      error('no such algo');
    end
    boundFunc = [];

  case {'logitGP'}
    switch algo
    case 'var-log'
      inferFunc = @inferMultiLogitGP_log;
    case 'var-boh'
      inferFunc = @inferMultiLogitGP_bohning;
      if isempty(options.boundParams)
        % form sA matrix (square root of A)
        M = params.others.nClass-1;
        N = size(data.X,1);
        sA = sparse([]);
        A1 = (eye(M) - ones(M,M)/(M + 1))/2;
        sA1 = chol(A1);
        for n = 1:N
          sA = blkdiag(sA,sA1);
        end
        options.boundParams.sA = sA;
        options.boundParams.A = sA'*sA;
      end
    case 'var-pw'  
      inferFunc = @inferStickLogitGP_pw;
      options.ind = data.ind;
      if isempty(options.boundParams)
        options.boundParams = getPiecewiseBound('quad',nPieces);
      end
    case 'ep'

    otherwise 
      error('no such algo');
    end
    boundFunc = [];

  otherwise
    error('no such model');
  end

