function [ss, logLik, postDist] = inferMultiLogitLGM_bohning(data, params, funObj, options)
% Inference on multiLogitLGM model minimizing variational free energy using 
% the Bohning bound.
%
% Written by Emtiyaz, CS, UBC
% Modified on Dec. 13, 2011

  [maxItersInfer, lowerBoundTol, initVec, display, computeSs, estimateMean, estimateCovMat, estimateLoadingMat, estimateBias, boundParams] = myProcessOptions(options, 'maxItersInfer', 30, 'lowerBoundTol', 1e-6, 'initVec', [], 'display', 1, 'computeSs', 0, 'estimateMean', 0, 'estimateCovMat', 0, 'estimateLoadingMat', 0, 'estimateBias', 0, 'boundParams', []);

  % get data
  Y = data.Y;
  Yunique = data.Yunique;
  wY = data.Ycount;
  Ybitvec = data.Ybitvec;
  miss = data.missing;
  J = data.mappingJ;
  I = data.mappingI;

  % get D,N,L,K
  [D,N] = size(Y);
  Nbitvec = size(Ybitvec,2);
  L=size(params.mean,1);
  K = params.others.nClass;
  nC = K-1;
  Widx = params.others.loadingMatIdx;

  % YbitVec, remove the baseline categories
  Ybitvec = Ybitvec(Widx,:);

  % get parameters
  invSigma = params.precMat;
  logdetSigma = params.logDetCovMat;
  Mu = params.mean;
  W = params.loadingMat(Widx,:);% remove the baseline category
  w0 = params.bias(Widx,:);% remove the baseline category

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Following are some pre-computed quantities, these can be done outside 
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % unique #categories
  nCu = unique(nC);

  % compute the bound matrix A
  if isempty(boundParams)
    for d = 1:length(nCu)
      A{d} = (eye(nCu(d))- ones(nCu(d),nCu(d))/(nCu(d) + 1))/2;
      iA{d} = inv(A{d});
    end
  else
    A = boundParams.A;
    iA = boundParams.iA;
  end

  % unique missing pattern matrix (DxN)
  miss = isnan(Yunique);
  [obsDim, Im, Jm] = unique(~miss','rows');
  obsDim = obsDim'; 

  % unique missing pattern matrix (sum(nC)xN)
  miss = isnan(Ybitvec);
  [obs, Im, Jm] = unique(~miss','rows');
  obs = obs';

  % If nC = [2 4 3], then nCd = [2 2 4 4 4 4 3 3 3]
  nCd = zeros(sum(nC),1);
  for d = 1:D
    idx = sum(nC(1:d-1))+1:sum(nC(1:d));
    nCd(idx) = nC(d);
  end

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Posterior Covariance 
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % for each missing pattern, compute covariance matrix
  for n = 1:size(obsDim,2)
    Onbv = obs(:,n);        % observed indx in bitvec
    On = find(obsDim(:,n)); % observed dimensions

    % posterior covariance
    WAW = 0;
    for d = On(:)'
      dIdx = sum(nC(1:d-1))+1:sum(nC(1:d));
      Ad = A{find(nC(d)==nCu)};
      WAW = WAW + W(dIdx,:)'*Ad*W(dIdx,:);
    end
    Vn = inv(invSigma + WAW);
    V(:,:,n) = Vn;

    % pre-compute quantities for posterior mean
    VSigmaInvMu(:,n) = Vn*invSigma*Mu;
    VW{n} = Vn*W(Onbv,:)';
    % pre-compute quantities for log-lik computation
    logdetV(n) = logdet(Vn);
    trVinvSigma(n) = Vn(:)'*invSigma(:);
    trAWVW(n) = 0;
    for d = On(:)'
      Ad = A{find(nC(d)==nCu)};
      dIdx = sum(nC(1:d-1))+1:sum(nC(1:d));
      trAWVW(n) = trAWVW(n) + trace(Ad*W(dIdx,:)*Vn*W(dIdx,:)');
    end
  end
  V = V(:,:,Jm); % map to unique bit patterns

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Posterior Mean
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % initialize
  if isempty(initVec)
    M = repmat(rand(length(Mu),1), 1, Nbitvec);
  else
    M = initVec;
  end
  mbar = (~miss).*bsxfun(@plus, W*M, w0);
  mbar = full(mbar);

  % pre-compute A*w0
  Aw0 = zeros(sum(nC),1);
  for d = 1:D
    dIdx = sum(nC(1:d-1))+1:sum(nC(1:d));
    Ad = A{find(nC(d)==nCu)};
    Aw0(dIdx) = Ad*w0(dIdx);
  end

  % inference
  logLik_n = zeros(1,Nbitvec);
  for iter = 1:maxItersInfer
    logLik(iter) = 0;

    % update variational parameter
    Psi = mbar; % size sum(nC)x Nbitvec, zeros at missing value
    YplusB_all = zeros(size(mbar));

    % for each missing pattern
    for n = 1:size(obsDim,2)
      Onbv = obs(:,n);        % Observed indx in BitVec
      On = find(obsDim(:,n)); % Observed dimensions
      nIdx = find(Jm==n);     % measurements with this missing pattern
      nCn = nC(On(:));
      Dn = length(On(:));
      Nn = length(nIdx);

      % map the data and varParams to observed data
      Yn = Ybitvec(Onbv,nIdx);
      Psi_n = Psi(Onbv,nIdx);

      % initialize G, B, A*Psi and lse
      G = zeros(size(Psi_n));
      B = G;
      APsi_n = G;
      lse = zeros(Dn,Nn);

      % For each unique category, we find Psi_nk, vectorize it and then perform
      % parallel computation of lse, g, and b
      nCdo = nCd(Onbv(:)); % nC.*obs(Yndk)
      nCdu = unique(nCdo);    
      for k = 1:length(nCdu)
        % find block of psi_nk for a unique category
        idx = find(nCdo == nCdu(k));  
        Dk = length(idx)/nCdu(k);
        Psi_nk = Psi_n(idx,:);
        Psi_nk = reshape(Psi_nk, nCdu(k), Dk*Nn);

        % perform parallel computation
        lsek = logsumexp([Psi_nk; zeros(1,Dk*Nn)],1); % lse
        Gk = exp(bsxfun(@minus, Psi_nk, lsek));       % g = exp(psi-lse)
        Ad = A{find(nCdu(k)==nCu)};                   
        APsi_nk = Ad*Psi_nk;                          
        Bk = APsi_nk(:) - Gk(:);                      % B = A*psi-g

        % reshape and fill
        G(idx,:) = reshape(Gk, nCdu(k)*Dk, Nn);
        APsi_n(idx,:) = reshape(APsi_nk, nCdu(k)*Dk, Nn);
        B(idx,:) = reshape(Bk, nCdu(k)*Dk, Nn);
        dIdx = find(nCn == nCdu(k));  
        lse(dIdx,:) = reshape(lsek, Dk, Nn);
      end

      % pseudo measurements
      YplusB = Yn + B;
      Ypseudo = bsxfun(@minus, YplusB, Aw0(Onbv));

      % store all of the Yplus for ss computation later
      YplusB_all(Onbv,nIdx) = YplusB;

      % compute posterior mean
      Mn = bsxfun(@plus, VW{n}*Ypseudo, VSigmaInvMu(:,n));
      M(:,nIdx) = Mn;

      % compute mbar
      mbar_n = bsxfun(@plus, W(Onbv,:)*Mn, w0(Onbv));
      mbar(Onbv,nIdx) = mbar_n;

      %% log likelihood, all quantities are of length Nn 
      % KL contribution
      diff = bsxfun(@minus, Mn, Mu);
      Dkl = 0.5*(logdetSigma-logdetV(n) + trVinvSigma(n) - L...
                  + sum(diff.*(invSigma'*diff),1));
      % contribution from the bound with parallel computation of m'*A*m
      mAm = zeros(1,Nn); 
      for k = 1:length(nCdu)
        idx = find(nCdo == nCdu(k));  
        Dk = length(idx)/nCdu(k);
        mbar_nk = mbar_n(idx,:);
        mbar_nk = reshape(mbar_nk, nCdu(k), Dk*Nn);
        Ad = A{find(nCdu(k)==nCu)};                   
        mAmk = sum(mbar_nk.*(Ad*mbar_nk),1); % sum across k 
        mAm = mAm + sum(reshape(mAmk,Dk,Nn),1); % sum across d
      end
      c = 0.5*sum(Psi_n.*APsi_n,1) - sum(G.*Psi_n,1) + sum(lse,1);
      quadBound = 0.5*mAm - sum(B.*mbar_n) + c + 0.5*trAWVW(n); 
      % final loglik
      logLikn = -Dkl + sum(Yn.*mbar_n) - quadBound;
      logLik_n(nIdx) = logLikn;
      logLik(iter) = logLik(iter) + N*sum(wY(nIdx).*logLikn);
    end

    % convergence
    [converged, incr] = isConverged(logLik, 1e-4, 'objFun');
    if display; fprintf('iter %d logLik %f\n', iter, logLik(iter)); end
    if converged; break; end;
    if i == maxItersInfer; warning('Maximum iterations reached'); end
    if incr < -1e-10; warning('Lower bound decreased'); end
  end
  meanPost = M;
  covMatPost = V;
  logLik = logLik(end);

  % postDist
  if nargout > 2
    postDist.mean = meanPost;
    postDist.covMat = covMatPost;
    postDist.logLik_n = logLik_n;
  end

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % sufficient statistics
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  if computeSs
    wMeanPost = bsxfun(@times, meanPost, wY);
    weightMat = permute(repmat(wY,[L,1,L]), [1 3 2]);
    wCovMatPost= covMatPost.*weightMat;
    if estimateMean
     ss.mean = sum(wMeanPost,2); 
      if estimateCovMat
        ss.covMat = sum(wCovMatPost,3) + wMeanPost*meanPost';
      end
    else
      if estimateCovMat
        diff = bsxfun(@minus, meanPost, params.mean);
        wDiff = bsxfun(@times, diff, wY);
        ss.covMat = sum(wCovMatPost,3) + wDiff*diff';
      end
    end
    if estimateLoadingMat
      % if no missing values
      if ~sum(miss(:))
        ss.rhs = sum(wCovMatPost,3) + wMeanPost*meanPost';
        ss.sumMean = sum(wMeanPost,2); 
        for d = 1:D
          iAd = iA{find(nC(d)==nCu)};                   
          idx = sum(nC(1:d-1))+1:sum(nC(1:d));
          ss.iAYplusBm(idx,:) = iAd*(YplusB_all(idx,:)*wMeanPost');
          sumYplusBm = sum(bsxfun(@times, wY, YplusB_all(idx,:)),2);
          ss.iAsumYplusBm(idx,:) = iAd*sumYplusBm;
        end
      % otherwise
      else
        for d = 1:D
          Od = find(~isnan(Yunique(d,:))); % measurements where d'th dim is observed
          Nd = sum(wY(Od)*N); % number of those measurements
          % ss for Wd
          ss.rhs(:,:,d) = sum(wCovMatPost(:,:,Od),3) + wMeanPost(:,Od)*meanPost(:,Od)';
          ss.sumMean(:,d) = sum(wMeanPost(:,Od),2); 
          iAd = iA{find(nC(d)==nCu)};                   
          idx = sum(nC(1:d-1))+1:sum(nC(1:d));
          ss.iAYplusBm(idx,:) = iAd*(YplusB_all(idx,Od)*wMeanPost(:,Od)');
          % ss for w0d
          sumYplusBm = sum(bsxfun(@times, wY(Od), YplusB_all(idx,Od)),2);
          ss.iAsumYplusBm(idx,:) = iAd*sumYplusBm;
          ss.NoverNd(d) = N/Nd;
        end
      end
    end
    ss.initVec = M;
  else
    ss = 0;
  end

