function [ss, logLik, postDist] = inferMultiLogitLGM_gradient(data, params, funObjBound, options)
% Inference on multiLogitLGM model minimizing variational free energy using a
% gradient method. 
%
% Written by Emtiyaz, CS, UBC
% Modified on Dec. 5, 2011

  [maxItersInfer, lowerBoundTol, initVec, display, computeSs, estimateMean, estimateCovMat, estimateLoadingMat, estimateBias, boundParams] = myProcessOptions(options, 'maxItersInfer', 30, 'lowerBoundTol', 1e-6, 'initVec', [], 'display', 1, 'computeSs', 0, 'estimateMean', 0, 'estimateCovMat', 0, 'estimateLoadingMat', 0, 'estimateBias', 0, 'boundParams', []);

  % get data
  Y = data.Y;
  Yunique = data.Yunique;
  wY = data.Ycount;
  Ybitvec = data.Ybitvec;
  miss = data.missing;
  J = data.mappingJ;
  I = data.mappingI;

  % get D,N,L,K
  [D,N] = size(Y);
  Nbitvec = size(Ybitvec,2);
  L=size(params.mean,1);
  K = params.others.nClass;

  % initialize
  if isempty(initVec)
    initVec = repmat([params.mean; packcovariance(params.covMat)], 1,Nbitvec);
  end

  % minfunc options
  optMinFunc.Display = display;
  optMinFunc.DerivativeCheck = 'off';
  optMinFunc.Method = 'lbfgs';
  optMinFunc.LS = 2;
  optMinFunc.MaxIter = maxItersInfer;
  optMinFunc.MaxFunEvals = maxItersInfer;
  optMinFunc.TolFun = lowerBoundTol;
  optMinFunc.TolX = lowerBoundTol;

  % inference
  logLik = 0;
  for n = 1:size(Yunique,2)
    % get indices for observed entries (and remove the baseline category)
    Widx = params.others.loadingMatIdx; % without baseline
    idx = find(~isnan(Ybitvec(:,n))); % observed
    obs = intersect(Widx,idx); % intersection of the two
    obsDim = find(~isnan(Yunique(:,n))); % obs Dimensions

    % form n'th data vector
    Yn = Ybitvec(obs,n);

    % params with new W, w0 and other information
    params_n = params;
    params_n.loadingMat = params.loadingMat(obs,:);
    params_n.bias = params.bias(obs);
    params_n.others.loadingMatIdx = obs;
    params_n.others.nClass = params_n.others.nClass(obsDim);
    % run minFunc
    v0=initVec(:,n);
    [v,fn,exitFlag, inform]=minFunc(@funObj_multiLogitLGM_gradient, v0, optMinFunc, Yn, params_n, funObjBound, boundParams);

    % add likelihood
    logLik_n(n)= -fn;
    logLik = logLik - N*wY(n)*fn;

    % get mean and covMat
    meanPost(:,n)=v(1:L);
    covMatPost(:,:,n)=unpackcovariance(v((L+1):(L + L*(L+1)/2)),L);
    regBias = v((L+L*(L+1)/2)+1:end);% only for regression

    % store values to warmstart next inference run
    vAll(:,n) = v;
    fcount(n) = inform.funcCount;
  end

  % postDist
  if nargout > 2
    postDist.mean = meanPost;
    postDist.covMat = covMatPost;
    postDist.logLik_n = logLik_n;
    postDist.maxFuncEval = max(fcount);
    if ~isempty(regBias)
      postDist.regBias = regBias;
    end
  end

  % sufficient statistics
  if computeSs
    wMeanPost = bsxfun(@times, meanPost, wY);
    weightMat = permute(repmat(wY,[L,1,L]), [1 3 2]);
    wCovMatPost= covMatPost.*weightMat;
    if estimateMean
      ss.mean = sum(wMeanPost,2); 
      ss.covMat = sum(wCovMatPost,3) + wMeanPost*meanPost';
    else
      diff = bsxfun(@minus, meanPost, params.mean);
      wDiff = bsxfun(@times, diff, wY);
      ss.covMat = sum(wCovMatPost,3) + wDiff*diff';
    end
    ss.initVec = vAll;
    if estimateLoadingMat
      ss.meanPost = meanPost;
      ss.covMatPost = covMatPost;
    end
  else
    ss = 0;
  end

