function [data, params] = setupLGM(data, model, options)
% Converts data to LGM format (data, params) 
% It is recommended to pass nClass in the data through options variable
% For binary data, data = ~data + 1 (code '0' as 2nd category and '1' as 1st
% category)
%TODO add error messages for wrong size of mean_ and covMat
%
% Written by Emtiyaz, CS, UBC
% Modified on Nov. 24, 2011

  switch model
  case {'bayesLogitReg','bayesMultiLogitReg'}
    % get user options if any
    [priorCovMat, priorMean, nClass] = myProcessOptions(options, 'priorCovMat', [], 'priorMean', [], 'nClass', []);
    % data
    X = data(:,1:end-1);
    y = data(:,end);
    clear data;

    % get K,M,N,D,F,L
    if isempty(nClass)
      K = length(unique(y));
    else
      K = nClass;
    end
    baseline = K;
    M = K-1;
    N = 1;              % 1 data vector in LGM
    D = length(y);      % dimension in LGM
    F = size(X,2);       
    L = F*M;        % latent dimension in LGM

    % Format Data for LGM
    data.Y = y; % data
    data.Yunique = y; % unique patterns, there is only one
    data.mappingJ = 1; % to map back from unique patterns to full data
    data.mappingI = 1; % to map back from unique patterns to full data
    data.Ycount = 1; % count of those patterns
    Ybitvec = full(bsxfun(@eq, y, [1:K])');
    data.Ybitvec = Ybitvec(:);% one of K encoding
    data.missing = isnan(y); % missing value pattern
    data.ind = data.Ybitvec;

    % indicators for piecewise bound
    Y11 = Ybitvec([end:-1:1],:);
    Y11 = cumsum(Y11);
    ind = Y11([end:-1:1],:);
    data.ind = ind(:);

    % prior mean
    if isempty(priorMean)
      params.mean = zeros(L,1);%mean_;
    else
      if length(priorMean)~=L error('priorMean should be of length #features*(#categories-1)'); end
      params.mean = priorMean;
    end
    
    % prior precMat and covMat 
    if isempty(priorCovMat)
      params.covMat = 1*eye(M*F);
    else
      [A1,A2] = size(priorCovMat);
      if A1~=A2; error('priorCovMat should be square'); end
      if A1~=L; error('priorCovMat should be a square matrix of dimension #features*(#categories-1)'); end
      if ~isposdef(priorCovMat); error('priorCovMat should posdef'); end;
      params.covMat = priorCovMat;
    end
    params.precMat = inv(params.covMat);
    params.logDetCovMat = logdet(params.covMat);

    % factor loading matrix and bias
    W = zeros(D*(M+1),F*M);
    for k = 1:K-1
      idx1 = k:K:(D-1)*K+k;
      idx2 = F*(k-1)+1:F*k;
      W(idx1,idx2) = X;
    end
    params.loadingMat = W;
    params.bias = zeros(size(W,1),1);

    % other information
    params.others.nClass = K*ones(D,1);
    v = [ones(K-1,1); 0];
    v = repmat(v, 1, D); 
    idx = find(v(:));
    params.others.loadingMatIdx = idx;

  case {'binaryLogitLGGM', 'binaryLogitFA', 'multiLogitLGGM','multiLogitFA','stickLogitLGGM','stickLogitFA'}

    % get user options if any
    [nClass, latentDim, loadingMat, bias, covMat, mean_] = myProcessOptions(options, 'nClass', [], 'latentDim', 2, 'loadingMat', [], 'bias', [], 'covMat', [], 'mean_', []);

    Y = data;
    clear data;

    % get K,M,N,D,F
    if isempty(nClass)
      K = max(Y,[],2);
    else
      K = nClass;
    end
    baseline = K;
    M = K-1;
    N = size(Y,2);
    D = size(Y,1);

    % latent dimension L
    switch model
    case {'binaryLogitLGGM','multiLogitLGGM','stickLogitLGGM'}
      L = sum(M);
    case {'binaryLogitFA','multiLogitFA','stickLogitFA'}
      L = latentDim;
    end

    % Format Data for LGM
    data.Y = Y; % data
    [Yunique,I,J]  = unique(Y','rows'); % unique patterns
    wY = hist(J,1:size(Yunique,1))/N;
    data.Yunique = Yunique'; 
    data.mappingJ = J; % to map back from unique patterns to full data
    data.mappingI = I; % to map from full data to unique patterns 
    data.Ycount = wY; % count of those patterns
    % TODO optimize this
    data.Ybitvec = encodeDataOneOfM(Yunique', K, 'M');% one of K encoding
    data.missing = isnan(Y); % missing value pattern

    % indicators for piecewise bound
    switch model
    case {'binaryLogitLGGM', 'binaryLogitFA'}
      Ybitvec = data.Ybitvec;
      Y11 = reshape(Ybitvec,2,D*size(Ybitvec,2));
      Y11 = Y11([end:-1:1],:);
      Y11 = cumsum(Y11);
      ind = Y11([end:-1:1],:);
      ind = reshape(ind,2*D,size(Ybitvec,2));
      data.ind = ind;

    case {'stickLogitLGGM','stickLogitFA'}
    % TODO optimize this
      data.ind = zeros(size(data.Ybitvec)); % indicator for I(k<=y_n)
      for d = 1:D
        idx = sum(K(1:d-1))+1:sum(K(1:d));
        Yd = data.Ybitvec(idx,:);
        miss = isnan(Yd);
        Yd(miss) = 0; % set missing entries to zero
        ind = ~cumsum(Yd,1) + Yd;
        ind(miss) = 0; % set missing entries to zero
        data.ind(idx,:) = ind;
      end
      data.ind = data.ind;
    end

    % prior mean, precMat, and covMat 
    if isempty(mean_)
      mean_ = zeros(L,1);
    end
    params.mean = mean_;
    if isempty(covMat)
      covMat = eye(L);
    end
    params.covMat = covMat;
    params.precMat = inv(covMat);
    params.logDetCovMat = logdet(covMat);

    % indexes for baseline category
    idx = ones(sum(K),1);
    sK = cumsum(K);
    idx([0; sK(1:end-1)] + baseline(:)) = 0;
    idx = find(idx);

    % factor loading matrix and bias
    if isempty(loadingMat)
      switch model
      case {'binaryLogitLGGM','multiLogitLGGM','stickLogitLGGM'}
        loadingMat = eye(sum(K),sum(K));
        loadingMat = loadingMat(:,idx);
      case {'binaryLogitFA','multiLogitFA','stickLogitFA'}
        loadingMat = zeros(sum(K),L);
        loadingMat(idx,:) = randn(sum(K-1),L);
      end
    end
    params.loadingMat = loadingMat;
    if isempty(bias)
      params.bias = zeros(sum(K),1);
    end

    % hyperparameters
    params.hyp.S0 = zeros(L,L);
    params.hyp.nu0 = -(L+1);

    % other information
    params.others.nClass = K;
    params.others.latentDim = L;
    params.others.loadingMatIdx = idx;
    params.others.nDim = D;
    params.others.nObs = N;

  case {'logitGP', 'multiLogitGP','stickLogitGP'}
    [meanFunc, hypMean, covFunc, hypCov, nClass] = myProcessOptions(options, 'meanFunc', {{@meanZero}}, 'hypMean', [], 'covFunc', {{@covSEiso}}, 'hypCov', [], 'nClass', []);

    X = data(:,1:end-1);
    y = data(:,end);
    N = length(y);
    clear data;

    % nClass
    if isempty(nClass)
      K = length(unique(y));
    else
      K = nClass;
    end
    M = K-1;

    % data
    data.y = y;
    data.X = X;
    ybitvec = bsxfun(@eq, y, [1:M])';
    data.ybitvec = ybitvec(:);

    % indicator for stick likelihood I(k<=y_n), vector of length M*N
    ind = reshape(ybitvec, M, N);
    ind = ~cumsum(ind,1) + ind;
    data.ind = ind(:);

    % params
    params.meanfunc = meanFunc{:};
    if ~isempty(hypMean)
      params.hyp.mean = hypMean;
    else
      params.hyp.mean = [];
    end
    params.covfunc = covFunc{:};
    if ~isempty(hypCov)
      params.hyp.cov = hypCov;
    else
      params.hyp.cov = 1*ones(M,2);
    end
    params.others.nClass = K;

  otherwise 
    error('no such model');
  end

